{-#  OPTIONS -fglasgow-exts  #-}
module LazyContract (

  -- contracts
  Contract(..), P(..), assert, demand, Box,

  -- blame assignment
  unitLocs, Fun(..),

  -- testing
  isContractViolation
) where

import TypeMap

import Data.IORef
import Control.Monad.State
import qualified Control.Exception as Exception
import qualified Data.List as List
import System.IO.Unsafe ( unsafePerformIO )

------------------------------------------------------------------------------
--                      CONTRACTS                                           --
------------------------------------------------------------------------------

--
-- Type definitions
--

-- The 'Contract' type is based on the types for contracts in the Hinze paper.
data Contract :: * -> * where
  Prop     :: (Box a a -> P a a) -> Contract a
  Function :: Contract a -> (Box a a -> Contract b) -> Contract (Fun a b)  
  Pair     :: Contract a -> Contract b -> Contract (a,b)
  Sum      :: Contract a -> Contract b -> Contract (Either a b)
  And      :: Contract a -> Contract a -> Contract a

data P :: * -> * -> * where
  CasePair :: Box t (a,b) -> ((Box t a, Box t b) -> P t c) -> P t (a,b)
  CaseSum  :: Box t (Either a b) -> (Box t a -> P t c) -> (Box t b -> P t d) 
           -> P t (Either a b)
  Pred     :: String -> CM Bool -> P t a

newtype Box t a = Box (IORef (Maybe a), Addr (Box t) t (Box t a))

instance Eq (Box t a) where
  Box (r, _) == Box (s, _) = r == s

type CStore = IORef [IO EnforceResult]

data EnforceResult = TryAgain | OK | Blame String
  deriving (Show,Eq)

-- The monad for contract predicates
newtype CM a = CM (IO a) deriving (Monad)

-- The type P' is similar to P but additionally carries the boxes that should be used
-- for the instrumentation.
data P' :: * -> * -> * where
  CasePair' :: Box t (a,b) -> Box t a -> Box t b -> ((Box t a, Box t b) -> P' t c)
            -> P' t (a,b)
  CaseSum'  :: (Box t (Either a b)) -> Box t a -> Box t b 
            -> (Box t a -> P' t c) -> (Box t b -> P' t d) -> P' t (Either a b)
  Pred'     :: String -> CM Bool -> P' t a


--
-- Operations
--

-- Get the address of the box
getAddr :: Box t a -> Addr (Box t) t (Box t a)
getAddr (Box (_, s)) = s

-- Creates the boxes necessary for the instrumentation.
createBoxes :: Map (Box a) a -> P a c -> IO (P' a c, Map (Box a) a)
createBoxes m (Pred s p) =
  return (Pred' s p, m)
createBoxes m (CasePair b f) =
  let
    myaddr = getAddr b
    mkfstaddr = addFst myaddr
    mksndaddr = addSnd myaddr
    fstaddr = mkfstaddr undefined
    sndaddr = mksndaddr undefined
  in
  do (b1, m1) <-
           case lkup fstaddr m of 
	     Just b0 -> return (b0, m)
	     Nothing -> do b0 <- freshBox mkfstaddr
			   return (b0, insert (getAddr b0) m)
     (b2, m2) <- 
           case lkup sndaddr m1 of
	     Just b0 -> return (b0, m1)
	     Nothing -> do b0 <- freshBox mksndaddr
			   return (b0, insert (getAddr b0) m1)
     (p, m3) <- createBoxes m2 (f (b1,b2))
     return (CasePair' b b1 b2 (\_ -> p), m3)
createBoxes m (CaseSum b f1 f2) =
  let
    myaddr = getAddr b 
    mklftaddr = addLft myaddr
    mkrgtaddr = addRgt myaddr
    lftaddr = mklftaddr undefined
    rgtaddr = mkrgtaddr undefined
  in
  do (b1, m1) <- 
           case lkup lftaddr m of 
	     Just b0 -> return (b0, m)
	     Nothing -> do b0 <- freshBox mklftaddr
			   return (b0, insert (getAddr b0) m)
     (b2, m2) <-
           case lkup rgtaddr m1 of
	     Just b0 -> return (b0, m1)
	     Nothing -> do b0 <- freshBox mkrgtaddr
			   return (b0, insert (getAddr b0) m1)
     (p1, m3) <- createBoxes m2 (f1 b1)
     (p2, m4) <- createBoxes m3 (f2 b2)
     return (CaseSum' b b1 b2 (\_ -> p1) (\_ -> p2), m4)

-- Predicate extraction. Corresponds to the \mathcal P translation of the
-- formalization.
extractPred :: Locs -> P' t a -> IO EnforceResult
extractPred ls p =
  Exception.catchJust isNotDemandable (extractPred' p) (\_ -> return TryAgain)
  where extractPred' :: P' t a -> IO EnforceResult
        extractPred' (Pred' s cm) =
          do b <- runCM cm
             return $ if b then OK else Blame ("predicate " ++ s ++ " failed, "
                                               ++ blame ls)
        extractPred' (CasePair' b b1 b2 f) =
          do _ <- runCM $ demand b -- really necessary?
             extractPred' (f (b1,b2))
        extractPred' (CaseSum' b b1 b2 f1 f2) =
          do x <- runCM $ demand b
             case x of
               Left _ -> extractPred' (f1 b1)
               Right _ -> extractPred' (f2 b2)

instrument :: CStore -> Map (Box t) t -> t -> t
instrument cstore m x =
  instr m x
  where
    instr :: Map (Box t) a -> a -> a
    instr PlugNode x = x
    instr (ValNode b m) x = unsafePerformIO $ do fill (Just b) x
						 checkCStore cstore
						 return (instr m x)
    instr (PairNode ma mb) (x1, x2) = (instr ma x1 ,instr mb x2)
    instr (SumNode ma mb) x = case x of
				Left x1 -> Left (instr ma x1)
				Right x2 -> Right (instr mb x2)


-- The assert function. Corresponds to the translation \mathcal C_l in
-- the formalization.
assert :: Contract a -> Locs -> a -> a
assert ls x = assert' Nothing store ls x
  where store = unsafePerformIO (newIORef [])
{-# NOINLINE assert #-}

assert' :: Maybe (Box t a) -> CStore -> Contract a -> Locs -> a -> a
assert' mbox cstore (Prop f) ls x = unsafePerformIO $
  do b <- freshBox AtTop
     let m = insert (getAddr b) emptyMap
     (p', m') <- createBoxes m (f b)
     addToCStore cstore (extractPred ls p')
     fill (Just b) x
     fill mbox x
     checkCStore cstore
     return (instrument cstore m' x)
assert' mbox store (Function c f) ls g = unsafePerformIO $
  do fill mbox g
     checkCStore store
     return $ Fun (\lx x -> unsafePerformIO $
                     do b <- freshBox AtTop
                        return (((assert' Nothing store (f b) ls) . 
                                 (app g lx) .
                                 (assert' (Just b) store c (ls `combineLocs` lx))) x))
assert' mbox store (Pair c1 c2) ls p = unsafePerformIO $
  do fill mbox p
     checkCStore store
     let  p' = (assert' Nothing store c1 ls (fst p),
                assert' Nothing store c2 ls (snd p))
     return p'
assert' mbox store (Sum c1 c2) ls s = unsafePerformIO $
  do fill mbox s
     checkCStore store
     let s' = case s of
                Left x -> Left (assert' Nothing store c1 ls x)
                Right x -> Right (assert' Nothing store c2 ls x)
     return s'
assert' mbox store (And c1 c2) ls x = 
  (assert' mbox store c1 ls . assert' mbox store c2 ls) x
{-# NOINLINE assert' #-}

--
-- Auxiliaries
--
isBlame (Blame _) = True
isBlame _         = False

isTryAgain TryAgain = True
isTryAgain _        = False

violationPrefix = "contract violation: "

isContractViolation :: Exception.Exception -> Maybe String
isContractViolation (Exception.ErrorCall s) =
  let (pref, suf) = splitAt (length violationPrefix) s
  in if pref == violationPrefix then Just suf else Nothing
isContractViolation _ = Nothing

checkCStore :: CStore -> IO ()
checkCStore store =
  do l <- readIORef store
     results <- sequence l
     case List.find isBlame results of
       Just (Blame s) -> error (violationPrefix ++ s)
       Nothing ->
         let l' = filter (\ (_, r) -> isTryAgain r) (zip l results)
         in writeIORef store (fst $ unzip l')

addToCStore :: CStore -> IO EnforceResult -> IO ()
addToCStore cstore x =
  do l <- readIORef cstore
     writeIORef cstore (x:l)

notDemandableMsg = "not-demandable"

isNotDemandable :: Exception.Exception -> Maybe ()
isNotDemandable (Exception.ErrorCall s) | s == notDemandableMsg = Just ()
                                        | otherwise             = Nothing
isNotDemandable _ = Nothing

demand :: Box t a -> CM a
demand (Box (ref,_)) = CM $
  do v <- readIORef ref
     case v of
       Nothing -> 
         do error notDemandableMsg
       Just x -> return x

fill :: Maybe (Box t a) -> a -> IO ()
fill (Just (Box (ref, _))) x = x `seq` -- important!
  do writeIORef ref (Just x)
fill Nothing _ = return ()

freshBox :: (Box t a -> Addr (Box t) t (Box t a)) -> IO (Box t a)
freshBox mkaddr = do ref <- newIORef Nothing
		     let box = Box (ref, addr)
			 addr = mkaddr box
		     return box

runCM (CM io) = io


------------------------------------------------------------------------------
--                      BLAME ASSIGNMENT                                    --
------------------------------------------------------------------------------

-- (taken from the Hinze paper)

data Locs = Pos Loc | NegPos Loc Loc
type Loc = String

unitLocs :: Loc -> Locs
unitLocs s = Pos s

combineLocs :: Locs -> Locs -> Locs
combineLocs (Pos l) (Pos l') = NegPos l l'
combineLocs (NegPos l' l) _ = NegPos l l'

blame :: Locs -> String
blame l =
  let s = case l of
            Pos x -> x
            NegPos _ x -> x
    in "the expression " ++ show s ++ " is to blame"

newtype Fun a b = Fun { app :: Locs -> a -> b }

